;;;; chicken-record.scm


(declare
  (fixnum)
  (hide index) )

(define (index x lst)
  (let loop ([i 0] [lst lst])
    (cond [(null? lst) (error "field not found" x)]
	  [(eq? x (car lst)) i]
	  [else (loop (add1 i) (cdr lst))] ) ) )

(define (make-record-type tname fields)
  (##sys#check-string tname 'make-record-type)
  (##sys#check-list fields 'make-record-type)
  (##sys#make-structure 'rtd (string->symbol tname) fields (length fields)) )

(define (record-type-descriptor rec)
  (##sys#block-ref rec 1) )

(define (record-constructor rtd #!optional fields)
  (##sys#check-structure rtd 'rtd 'record-constructor)
  (let ([rfields (##sys#slot rtd 2)]
	[id (##sys#slot rtd 1)] 
	[len (##sys#slot rtd 3)] )
    (if (list? fields)
	(let ([n (length fields)])
	  (lambda args0
	    (let ([args (list->vector args0)])
	      (unless (= (##sys#size args) n)
		(error "wrong number of arguments to record constructor" args0 id) )
	      (apply
	       ##sys#make-structure
	       id
	       rtd
	       (map (lambda (f) (vector-ref args (index f rfields))) fields) ) ) ) )
	(lambda args
	  (let ([x (apply ##sys#make-structure id rtd args)])
	    (unless (= (- (##sys#size x) 2) len)
	      (error "wrong number of arguments to record constructor" args id) )
	    x) ) ) ) )

(define (record-predicate rtd)
  (##sys#check-structure rtd 'rtd 'record-predicate)
  (lambda (x) 
    (and (##sys#generic-structure? x)
	 (eq? rtd (##sys#slot x 1)) ) ) )

(define (record-accessor rtd fname)
  (##sys#check-structure rtd 'rtd 'record-accessor)
  (##sys#check-symbol fname 'record-accessor)
  (let ([i (+ 2 (index fname (##sys#slot rtd 2)))]
	[id (##sys#slot rtd 1)] )
    (lambda (x)
      (##sys#check-structure x id)
      (##sys#slot x i) ) ) )

(define (record-modifier rtd fname)
  (##sys#check-structure rtd 'rtd 'record-modifier)
  (##sys#check-symbol fname 'record-modifier)
  (let ([i (+ 2 (index fname (##sys#slot rtd 2)))]
	[id (##sys#slot rtd 1)] )
    (lambda (x y)
      (##sys#check-structure x id)
      (##sys#setslot x i y) ) ) )
