#| -*-Scheme-*-

$Id: syntax-output.scm,v 14.10 2003/03/14 01:12:39 cph Exp $

Copyright 1989,1990,1991,2001,2002,2003 Massachusetts Institute of Technology

This file is part of MIT/GNU Scheme.

MIT/GNU Scheme is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

MIT/GNU Scheme is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with MIT/GNU Scheme; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

|#

;;;; Syntaxer Output Interface

(declare (usual-integrations))

(define (syntax-error history . rest)
  history				;ignore
  (apply error rest))

(define (transformer-eval expression environment)
  (eval expression environment))

(define (output/variable name)
  name)

(define (output/constant datum)
  datum)

(define (output/assignment name value)
  `(set! ,name ,value) )

(define (output/top-level-definition name value)
  `(define ,name ,value) )

(define (output/top-level-syntax-definition name value)
  `(define-syntax ,name ,value) )

(define (output/conditional predicate consequent alternative)
  `(if ,predicate ,consequent ,alternative) )

(define (output/disjunction predicate alternative)
  `(or ,predicate ,alternative) )	;***???

(define (output/sequence expressions)
  `(begin ,@expressions) )

(define (output/combination operator operands)
  (cons operator operands) )

(define (output/lambda lambda-list body)
  `(lambda ,lambda-list ,body) )

(define (output/named-lambda name lambda-list body)
  `(lambda ,lambda-list ,body) )

(define (output/delay expression)
  `(delay ,expression) )

(define (output/unassigned-test name)
  (make-unassigned? name))

(define (output/unassigned)
  (make-unassigned-reference-trap))

(define (output/unspecific)
  '(void) )

(define (output/let names values body)
  (output/combination (output/named-lambda (gensym 'let) names body) values))

(define (output/letrec names values body)
  `(letrec ,(map list names values) ,@body) )

(define (output/body declarations body)
  `(let () ,body) )

(define (output/definition name value)
  `(define ,name ,value) )

(define (output/top-level-sequence declarations expressions)
  (output/sequence expressions) )

(define (output/the-environment)
  (make-the-environment))

(define (output/access-reference name environment)
  (make-access environment name))

(define (output/access-assignment name environment value)
  (make-combination lexical-assignment (list environment name value)))

(define (output/local-declare declarations body)
  (make-declaration declarations body))


;;; Post processing

(define (output/post-process-expression expression)
  expression
  #;  (let ((unmapping (empty-unmapping)))
  (compute-substitution expression unmapping)
  (alpha-substitute (unmapping->substitution unmapping) expression)))

(define (empty-unmapping)
  (make-hash-table eq?))

(define (store-unmapping-entry! identifier unmapped-identifier unmapping)
  (hash-table-set! unmapping identifier unmapped-identifier))

(define ((unmapping->substitution unmapping) identifier)
  (or (hash-table-ref unmapping identifier)
      (finalize-mapped-identifier identifier)))

(define (unmap-identifier identifier)
  (let ((entry
	 (hash-table-ref (rename-database/unmapping-table *rename-database*)
			 identifier
			 )))
    (if entry
	(identifier->symbol (car entry))
	(begin
	  (if (not (symbol? identifier))
	      (error "bad-range-argument" identifier 'UNMAP-IDENTIFIER))
	  identifier))))

(define (finalize-mapped-identifier identifier)
  (let ((entry
	 (hash-table-ref (rename-database/unmapping-table *rename-database*)
			 identifier
			 )))
    (if entry
	(let ((identifier (car entry))
	      (frame-number (force (cdr entry))))
	  (if (interned-symbol? identifier)
	      (map-interned-symbol identifier frame-number)
	      (map-uninterned-identifier identifier frame-number)))
	(begin
	  (if (not (symbol? identifier))
	      (error "bad-range-argument" identifier
					'FINALIZE-MAPPED-IDENTIFIER))
	  identifier))))

(define (map-interned-symbol symbol frame-number)
  (string->symbol
   (string-append "."
		  (symbol-name symbol)
		  "."
		  (number->string frame-number))))

(define (map-uninterned-identifier identifier frame-number)
  (let ((table (rename-database/id-table *rename-database*))
	(symbol (identifier->symbol identifier)))
    (let ((alist (hash-table-ref table symbol '())))
      (let ((entry (assv frame-number alist)))
	(if entry
	    (let ((entry* (assq identifier (cdr entry))))
	      (if entry*
		  (cdr entry*)
		  (let ((mapped-symbol
			 (map-indexed-symbol symbol
					     frame-number
					     (length (cdr entry)))))
		    (set-cdr! entry
			      (cons (cons identifier mapped-symbol)
				    (cdr entry)))
		    mapped-symbol)))
	    (let ((mapped-symbol (map-indexed-symbol symbol frame-number 0)))
	      (hash-table-set! table
			       symbol
			       (cons (list frame-number
					   (cons identifier mapped-symbol))
				     alist))
	      mapped-symbol))))))

(define (map-indexed-symbol symbol frame-number index-number)
  (string->symbol
   (string-append "."
		  (symbol-name symbol)
		  "."
		  (number->string frame-number)
		  "-"
		  (number->string index-number))))

;;;; Identifiers

(define *rename-database*)

(define-record rename-database frame-number mapping-table unmapping-table id-table)

(define (initial-rename-database)
  (make-rename-database 0 (make-hash-table equal?) (make-hash-table eq?) (make-hash-table eq?)) )

(define rename-database/frame-number rename-database-frame-number)
(define set-rename-database/frame-number! rename-database-frame-number-set!)
(define rename-database/mapping-table rename-database-mapping-table)
(define rename-database/unmapping-table rename-database-unmapping-table)
(define rename-database/id-table rename-database-id-table)

(define (make-rename-id)
  (delay
    (let ((n (+ (rename-database/frame-number *rename-database*) 1)))
      (set-rename-database/frame-number! *rename-database* n)
      n)))

(define (rename-identifier identifier rename-id)
  (let ((key (cons identifier rename-id))
	(mapping-table (rename-database/mapping-table *rename-database*)))
    (or (hash-table-ref mapping-table key #f)
	(let ((mapped-identifier
	       (string->uninterned-symbol
		(symbol->string (identifier->symbol identifier)))))
	  (hash-table-set! mapping-table key mapped-identifier)
	  (hash-table-set! (rename-database/unmapping-table *rename-database*)
			   mapped-identifier
			   key)
	  mapped-identifier))))

(define (rename-top-level-identifier identifier)
  (if (symbol? identifier)
      identifier
      (rename-identifier identifier (delay 0))))

(define (make-name-generator)
  (let ((id (make-rename-id)))
    (lambda (identifier)
      (rename-identifier identifier id))))

